/***************************************************************************
                          s_multi_loop.cpp  -  description
                             -------------------
    begin                : Mon Apr 15 2002
    copyright            : (C) 2002 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <malloc.h>

#include "externs.h"
#include "common.h"
#include "s_multi_loop.h"
#include "simfold.h"
#include "s_energy_matrix.h"

s_multi_loop::s_multi_loop (int *seq, int length)
// The constructor
{
    int i;
    sequence = seq;
    seqlen = length;
    this->V = NULL;
    
    index = (int*) malloc (sizeof(int)*length);
    int total_length = (length *(length+1))/2;
    index[0] = 0;
    for (i=1; i < length; i++)
        index[i] = index[i-1]+length-i+1;

    WM = new int [total_length];
    if (WM == NULL) giveup ("Cannot allocate memory", "s_multi_loop");
    for (i=0; i < total_length; i++) WM[i] = INF;                                    
}



s_multi_loop::~s_multi_loop ()
// The destructor
{
    delete [] index;
    delete [] WM;
}


void s_multi_loop::compute_energy_WM (int j)
// compute de MFE of a partial multi-loop closed at (i,j)
{
    int i;
    int tmp;

    for (i=j-TURN-1; i>=0; i--)
    {
        int ij = index[i]+j-i;
        int iplus1j = index[i+1]+j-i-1;
        int ijminus1 = index[i]+j-1-i;


        tmp = V->get_energy(i,j) +
                   AU_penalty (sequence[i], sequence[j]) +
                   misc.multi_helix_penalty;
        if (tmp < WM[ij])
          {
            WM[ij] = tmp;
          }
        tmp = V->get_energy(i+1,j) +
                  AU_penalty (sequence[i+1], sequence[j]) +
                  dangle_bot [sequence[j]]
                             [sequence[i+1]]
                             [sequence[i]] +
                  misc.multi_helix_penalty +
                  misc.multi_free_base_penalty;
        if (tmp < WM[ij])
        {
            WM[ij] = tmp;
        }

        tmp = V->get_energy(i,j-1) +
                  AU_penalty (sequence[i], sequence[j-1]) +
                  dangle_top [sequence [j-1]]
                             [sequence [i]]
                             [sequence [j]] +
                  misc.multi_helix_penalty +
                  misc.multi_free_base_penalty;
        if (tmp < WM[ij])
        {
            WM[ij] = tmp;
        }

        tmp = V->get_energy(i+1,j-1) +
                  AU_penalty (sequence[i+1], sequence[j-1]) +
                  dangle_bot [sequence[j-1]]
                             [sequence[i+1]]
                             [sequence[i]] +
                  dangle_top [sequence [j-1]]
                             [sequence [i+1]]
                             [sequence [j]] +
                  misc.multi_helix_penalty +
                  2*misc.multi_free_base_penalty;
        if (tmp < WM[ij])
        {
                WM[ij] = tmp;
        }

        tmp = WM[iplus1j] + misc.multi_free_base_penalty;            
        if (tmp < WM[ij])
        {
            WM[ij] = tmp;
        }

        tmp = WM[ijminus1] + misc.multi_free_base_penalty;
        if (tmp < WM[ij])
        {
            WM[ij] = tmp;
        }

        for (int k=i; k < j; k++)
        {
            int ik = index[i]+k-i;
            int kplus1j = index[k+1]+j-k-1;
            tmp = WM[ik] + WM[kplus1j];
            if (tmp < WM[ij])
              {
                WM[ij] = tmp;
              }
        }
    }
}



int s_multi_loop::compute_energy (int i, int j)
// compute the MFE of a multi-loop closed at (i,j)
{
    int min = INF, tmp, k;
    int iplus1k;
    int kplus1jminus1;
    int iplus2k;
    int kplus1jminus2;

    for (k = i+TURN+1; k <= j-TURN-2; k++)
    {
        iplus1k = index[i+1] + k -i-1;
        kplus1jminus1 = index[k+1] + j-1 -k-1;
        iplus2k = index[i+2] + k -i-2;
        kplus1jminus2 = index[k+1] + j-2 -k-1;
        
        tmp = WM[iplus1k] + WM[kplus1jminus1];
        if (tmp < min)
            min = tmp;
              
        tmp = WM[iplus2k] + WM[kplus1jminus1] + 
              dangle_top [sequence [i]]
              [sequence [j]]
              [sequence [i+1]] + 
              misc.multi_free_base_penalty;
        if (tmp < min)
            min = tmp;

        tmp = WM[iplus1k] + WM[kplus1jminus2] + 
              dangle_bot [sequence[i]]
              [sequence[j]]
              [sequence[j-1]] + 
              misc.multi_free_base_penalty;
        if (tmp < min)
            min = tmp;

        tmp = WM[iplus2k] + WM[kplus1jminus2] + 
              dangle_top [sequence [i]]
              [sequence [j]]
              [sequence [i+1]] +  
              dangle_bot [sequence[i]]
              [sequence[j]]
              [sequence[j-1]] + 
              2 * misc.multi_free_base_penalty;
        if (tmp < min)
          min = tmp;
    }    
                          
    min += misc.multi_helix_penalty + misc.multi_offset +
           AU_penalty (sequence[i], sequence[j]);
    return min;
}



